#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/debugfs.h>
#include <linux/uaccess.h>

MODULE_AUTHOR("Alessandro Barenghi");
MODULE_DESCRIPTION("DebugFS file example");
MODULE_LICENSE("GPL");

char myvalue;
int file_value;
struct dentry *dir, *file;
char internal_buffer[200];

static ssize_t my_read_file(struct file *file, char __user *userbuf,
                                size_t count, loff_t *ppos)
{
	return simple_read_from_buffer(userbuf, count, ppos, internal_buffer, 200);
} 

static ssize_t my_write_file(struct file *file, const char __user *buf,
                                size_t count, loff_t *ppos)
{       
        int copied=0;
	if(count > 200)
		return -EINVAL;
	copied = copy_from_user(internal_buffer, buf, count);
	if(copied >=0){
	   return count;
	}
	else {
	  return 0;
	}
}

static const struct file_operations my_file_ops = {
        .read = my_read_file,
        .write = my_write_file,
};


static int __init debugexample_module_init(void)
{

	/* Create a directory in DebugFS */
	dir = debugfs_create_dir("debug_directory", NULL);
	/* Create a file in DebugFS, using dir as the parent directory*/
	file = debugfs_create_file("debug_file", /* Octal! */ 0644, dir, &file_value, &my_file_ops);

	return 0;
}

static void __exit debugexample_module_exit(void)
{
	debugfs_remove(file);
	debugfs_remove(dir);
}

module_init(debugexample_module_init);
module_exit(debugexample_module_exit);
