#include <linux/module.h>	/* Needed by all modules */
#include <linux/kernel.h>	/* Needed for KERN_INFO */
#include <linux/init.h>		/* Needed for the macros */
#include <linux/types.h>	/* Holds dev_t */
#include <linux/fs.h>		/* holds register_chrdev */
#include <asm/uaccess.h>	/* for put_user */
#include <linux/miscdevice.h>

MODULE_AUTHOR("Alessandro Barenghi");
MODULE_DESCRIPTION("A mockup character device module");
MODULE_LICENSE("GPL");

#define DEVICE_NAME "mockup_char"	/* Dev name as it appears in /proc/devices   */
#define BUF_LEN 80		/* Max length of the message from the device */

static int Major;		/* Major number assigned to our device driver */
static int Device_Open = 0;	/* Is device open?  
				 * Used to prevent multiple access to device */
static char msg[BUF_LEN];	/* The msg the device will give when asked */
static char *msg_Ptr;

static ssize_t
device_write(struct file *filp, const char *buff, size_t len, loff_t * off)
{
      if (Device_Open)
          return -EBUSY;
       Device_Open++;
	printk(KERN_ALERT "Sorry, this operation isn't supported.\n");
       Device_Open--; 
	return -EINVAL;
}

static int device_open(struct inode *inode, struct file *file)
{
	static int hello_world_counter = 0;

	if (Device_Open)
		return -EBUSY;

	Device_Open++;
	/* prepare the message */
	sprintf(msg, "I already told you %d times Hello world!\n", hello_world_counter++);
	msg_Ptr = msg;
	try_module_get(THIS_MODULE); /* marks this module as in use*/

	return 0;
}

static ssize_t device_read(struct file *filp,	/* see include/linux/fs.h   */
			   char *buffer,	/* buffer to fill with data */
			   size_t length,	/* length of the buffer     */
			   loff_t * offset)
{
	/*
	 * Number of bytes actually written to the buffer 
	 */
	int bytes_read = 0;

	/*
	 * If we're at the end of the message, 
	 * return 0 signifying end of file 
	 */
	if (*msg_Ptr == 0)
		return 0;

	/* 
	 * Actually put the data into the buffer 
	 */
	while (length && *msg_Ptr) {

		/* 
		 * The buffer is in the user data segment, not the kernel 
		 * segment so "*" assignment won't work.  We have to use 
		 * put_user which copies data from the kernel data segment to
		 * the user data segment. 
		 */
		put_user(*(msg_Ptr++), buffer++);

		length--;
		bytes_read++;
	}

	/* 
	 * Most read functions return the number of bytes put into the buffer
	 */
	return bytes_read;
}

static int device_release(struct inode *inode, struct file *file)
{
	Device_Open--;		/* We're now ready for our next caller */

	/* 
	 * Decrement the usage count, or else once you opened the file, you'll
	 * never get get rid of the module. 
	 */
	module_put(THIS_MODULE); 

	return 0;
}

static struct file_operations fops = {
	.read = device_read,
	.write = device_write,
	.open = device_open,
	.release = device_release
};

static struct miscdevice my_misc_device = {
    .minor = MISC_DYNAMIC_MINOR,
    .name = "mydevice",
    .fops = &fops
};

static int __init startup(void)
{
        Major = misc_register(&my_misc_device);

	if (Major < 0) {
	  printk(KERN_ALERT "Registering char device failed with %d\n", Major);
	  return Major;
	}

	printk(KERN_INFO "I was assigned major number %d. To talk to\n", Major);
	return 0;
}

static void __exit shutdown(void)
{
	/* 
	 * Unregister the device 
	 */
        printk(KERN_INFO "Unplugging character device\n");
	misc_deregister(&my_misc_device);

}

module_init(startup);
module_exit(shutdown);
